/*
Teeny Win32 demo program.

Orion Sky Lawlor, olawlor@acm.org, 2006/04/18 (Public Domain)
*/
#undef UNICODE /* Use ASCII versions of Windows calls, not Unicode */
#define _CRT_SECURE_NO_DEPRECATE 1 /* don't whine about fopen, man... */

// Windows Header Files:
#define WIN32_LEAN_AND_MEAN		// Exclude rarely-used stuff from Windows headers
#include <windows.h>

#include <resources/resource.h> // Windows resource IDs

// C Runtime Header Files
#include <cstdio>
#include <string>
#include <vector>

/*
Windows libraries.  Should be listed on the command line, but just in case...

In the worst case you may need all of these:
ADD LINK32 kernel32.lib user32.lib gdi32.lib winspool.lib comdlg32.lib advapi32.lib shell32.lib ole32.lib oleaut32.lib uuid.lib odbc32.lib odbccp32.lib
But in practice the below pragmas seem to suffice.
*/
#pragma comment (lib, "kernel32.lib") 
#pragma comment (lib, "user32.lib") 
#pragma comment (lib, "gdi32.lib") 


/*Logging facility: "WT" logs stuff to a disk file (one of the few ways to debug!) */
FILE *errLog(void) {
	static FILE *ret=NULL;
	if (ret==NULL) {
		ret=std::fopen("win32err.log","w");
		if (ret==NULL) ret=stderr;
	}
	return ret;
}
#if 1
#  define WT(str) fprintf(errLog(),"%.3f\t"str"\n",0.0);fflush(errLog());
#  define WT1(str,a) fprintf(errLog(),"%.3f\t"str"\n",0.0,a);fflush(errLog());
#  define WT2(str,a,b) fprintf(errLog(),"%.3f\t"str"\n",0.0,a,b);fflush(errLog());
#  define WT3(str,a,b,c) fprintf(errLog(),"%.3f\t"str"\n",0.0,a,b,c);fflush(errLog());
#else
#  define WT(str) /*empty*/
#  define WT1(str,a) /*empty*/
#  define WT2(str,a,b) /*empty*/
#  define WT3(str,a,b,c) /*empty*/
#endif

// Error checking
void abort(const char *msg,int line=-1) {
	fprintf(errLog(),"Win32 Toolkit Aborting: %s (line %d)\n",msg,line);
	exit(27);
}
#define BAD(err) abort(err,__LINE__)
#define APICHECK(fn,args) do {if (!fn args) BAD(#fn); }while(0)

/* Function Prototypes */
int parseCommandLine(char *args);
LRESULT CALLBACK myWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK myAboutBox(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);

/********************** Actual Code Starts Here **************************/

const char *WindowClass_name="OSL_GUI_Wndproc1_0";

class myWindow 
{
public:
	static HINSTANCE hInst; // current application instance
	static int nCmdShow; // "show type" we were invoked with

	/* Static Init: Called before anything else starts running */
	static void staticInit(HINSTANCE hInst_,int nCmdShow_)
	{
		hInst=hInst_;
		nCmdShow=nCmdShow_;
		
		//Register our new "window class" so we can handle events.
		WNDCLASSEX wcex;

		wcex.cbSize = sizeof(WNDCLASSEX); 
		wcex.style			= CS_HREDRAW | CS_VREDRAW;
		wcex.lpfnWndProc	= (WNDPROC)myWndProc; /* procedure to call with events */
		wcex.cbClsExtra		= 0;
		wcex.cbWndExtra		= 0;
		wcex.hInstance		= hInst;
		wcex.hIcon			= LoadIcon(hInst, (LPCTSTR)MYMAIN_BIG_ICON);
		wcex.hCursor		= LoadCursor(NULL, IDC_ARROW); // (LPCTSTR)MYMAIN_CURSOR);
		wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
		wcex.lpszMenuName	= (LPCTSTR)MYMENU_BAR;
		wcex.lpszClassName	= WindowClass_name;
		wcex.hIconSm		= LoadIcon(hInst, (LPCTSTR)MYMAIN_SM_ICON);
		RegisterClassEx(&wcex);
	}

protected:
	HWND hWnd; //Window handle
	HACCEL hAccelTable; // Menu Accellerators
	//Inherited: Point size;
	
public:
	myWindow(void)
	{
		WT("Creating Window")
		hAccelTable = LoadAccelerators(hInst, (LPCTSTR)MYMENU_BAR);
		const char *szTitle="Stupid Default Window Name";
		hWnd = CreateWindow(WindowClass_name, szTitle, WS_OVERLAPPEDWINDOW,
		      CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInst, NULL);
		if (!hWnd) BAD("CreateWindow");
		SetWindowLongPtr(hWnd,GWLP_USERDATA,(LONG_PTR)this); //<- can't check return code
		ShowWindow(hWnd,nCmdShow); // <- windows start invisible, so must show them...
	}

/** Event handling **/
	// Wait for an event to happen.  Call this in a loop in the main program.
	virtual void waitEvent(void)
	{
		MSG msg;
		if (1==GetMessage(&msg, hWnd, 0, 0))
			handleMessage(msg);
	}
	// Called at the start of each windows message 
	void handleMessage(MSG &msg) {
		if (0!=TranslateAccelerator(hWnd, hAccelTable, &msg)) return;
		TranslateMessage(&msg);
		DispatchMessage(&msg); /* calls myWndProc, which calls classWndProc */
	}

	// "Window Procedure": the main place events get handled
	virtual LRESULT classWndProc(UINT message, WPARAM wParam, LPARAM lParam)
	{
		switch (message) 
		{
	/*Keyboard*/
		case WM_CHAR: case WM_SYSCHAR:
			handleChar((TCHAR)wParam,lParam);
			return 0;
		case WM_KEYDOWN: case WM_KEYUP:
		case WM_SYSKEYDOWN: case WM_SYSKEYUP:
			handleKey((int)wParam,lParam,message);
			MessageBox(NULL,"A key went down","Caption",0);
			return 0;
	/*Mouse*/
		case WM_LBUTTONDOWN: case WM_LBUTTONUP:
		case WM_MBUTTONDOWN: case WM_MBUTTONUP:
		case WM_RBUTTONDOWN: case WM_RBUTTONUP:
		case WM_MOUSEMOVE:
			handleMouse(LOWORD(lParam),HIWORD(lParam),(int)wParam,message);
			return 0;
	/*Menu*/
		case WM_COMMAND: WT("WM_COMMAND")
			switch (LOWORD(wParam)) {
			case MYMENU_EXIT: exit(0); break;
			case MYMENU_OPEN: 
				DialogBox(NULL, MAKEINTRESOURCE(MYDIALOG_ABOUT),
						NULL, (DLGPROC)myAboutBox);
				break;
			default: 
				MessageBox(NULL,"Er, I don't know what to do for that menu item.  Sorry...","Crap!",0);
				break;
			};
			//eventDest->handleEvent(Event(Event::MENU_EVENT,evtTime,modifiers,
			//	LOWORD(wParam),0));
			return 0;
	/*Window*/
		case WM_ERASEBKGND: WT("WM_ERASEBKGND")
			/*Just ignore this, which prevents flashing on redraws*/
			return 1;
		case WM_PAINT:
			windowRepaint();
			//eventDest->handleEvent(Event(Event::WINDOW_REPAINT,evtTime,Event::NONE_MASK));
			return 0;
		case WM_CLOSE: WT("WM_CLOSE") //Asked to close window
			exit(0); /* Just exit normally */
			//eventDest->handleEvent(Event(Event::WINDOW_CLOSE,evtTime,modifiers));
			return 0;
		case WM_WINDOWPOSCHANGED: case WM_MOVING: {
			WT("WM_WINDOWCHANGE")
			/*getWindowPlacement();
			Event e(Event::WINDOW_MOVE,evtTime,Event::NONE_MASK,topLeft,0);
			eventDest->handleEvent(e);*/
			} break;
		//case WM_SIZING:
		case WM_SIZE:
			WT("WM_SIZE") {
			/*getWindowPlacement();
			Event e(Event::WINDOW_RESIZE,evtTime,Event::NONE_MASK,size,0);
			eventDest->handleEvent(e);*/
			}
			break;
		case WM_ACTIVATE:
			//eventDest->handleEvent(Event(wmActivate(LOWORD(wParam)),evtTime,modifiers));
			break;
		default:
			WT1("                      (unknown 0x%04x)",message)
			break;
		}
		/* Pass everything on to the default window handling */
		return DefWindowProc(hWnd, message, wParam, lParam);
	}

	/* Called to draw window from scratch */
	virtual void windowRepaint(void) 
	{
		/* Every redraw has to be wrapped in BeginPaint/EndPaint, or
		else Windows doesn't think it actually happened... */
		PAINTSTRUCT ps;
		HDC hdc=BeginPaint(hWnd, &ps);

		/* Clear the interior of the window */
		RECT rt;
		APICHECK(GetClientRect,(hWnd, &rt));
		FillRect(hdc, &rt, (HBRUSH)(COLOR_WINDOW+1));

		/* Draw some text in the window */
		char str[1000];
		sprintf(str,"Hello!");
		DrawText(hdc, str, (int)strlen(str), &rt, DT_CENTER);

		if (0) PaintDesktop(hdc);

		EndPaint(hWnd, &ps);
	}
	/* A key went down or came up */
	void handleKey(int winVK,LPARAM lp,UINT msg) 
	{
		int winOEM=(int)((lp>>16)&0x1ff);
		WT2("handleKey msg=%d winVK=%d",msg,winVK)
		/*
		int key=winVK2oslVK(winVK,0);
		eventDest->handleEvent(Event(t,evtTime,modifiers,
			winOEM,key));*/
	}
	/* A character was typed */
	void handleChar(TCHAR ch,LPARAM lp)
	{
		int repeatCount=(int)(lp&0xff);
		int winOEM=(int)((lp>>16)&0x1ff);
		int unicode=(int)ch;/*Unicode character*/
		WT2("handleChar winVK=%d unicode=%d",winOEM,unicode)
		/*
		int key=winVK2oslVK(winVK,unicode);
		for (int i=0;i<repeatCount;i++)
		  eventDest->handleEvent(Event(Event::KEY_TYPE,evtTime,modifiers,
			winOEM,key));*/
	}
	/* The mouse was moved or pressed */
	void handleMouse(int x,int y,int keys,UINT msg) 
	{
		WT3("handleMouse t=%02x (%d,%d)",msg,x,y)
		int buttonMask=0;
		if (keys & MK_LBUTTON) buttonMask|=(1<<0);
		if (keys & MK_MBUTTON) buttonMask|=(1<<1);
		if (keys & MK_RBUTTON) buttonMask|=(1<<2);
		/*
		Point p(x,y);
		Event::type_t t=Event::MOUSE_PRESS;
		if (msg==WM_LBUTTONUP || msg==WM_MBUTTONUP || msg==WM_RBUTTONUP)
			t=Event::MOUSE_RELEASE;
		else if (msg==WM_MOUSEMOVE) {
			if (buttonMask==0) t=Event::MOUSE_MOVE;
			else t=Event::MOUSE_DRAG;
		}
		eventDest->handleEvent(Event(t,evtTime,modifiers,p,buttonMask));
		*/
	}

	/* Destructor */
	~myWindow() {
		WT("Destroying Window")
		SetWindowLongPtr(hWnd,GWL_USERDATA,(LONG_PTR)NULL);
		DestroyWindow(hWnd);
	}

#if 0  /* Unused/demo code */
// Check if an event happened.  Immediately returns false if nothing happened.
	virtual bool checkEvent(void)
	{
		MSG msg;
		if (1==PeekMessage(&msg, hWnd, 0, 0, PM_REMOVE)) {
			handleMessage(msg);
			return true;
		} 
		else /* nothing happened */
			return false;
	}

//Draw this image at this (x,y) location
	virtual void draw(const ::osl::graphics2d::Raster &r,int x,int y)
	{
		if (r.wid==0 && r.ht==0) return; //Windows doesn't handle zero-size images
		HDC hdc = GetDC(hWnd); 
		if (!hdc) BAD("GetDC");
		const ::osl::graphics2d::RgbaRaster *rgb=
			dynamic_cast<const osl::graphics2d::RgbaRaster *>(&r);
		
		if (rgb!=NULL) 
		{ //Copy 24-bit data directly
			const void *data=(const void *)&(rgb->at(0,0));
			BITMAPV4HEADER hSto;
			BITMAPV4HEADER *h=&hSto;
			h->bV4Size=sizeof(*h);
			h->bV4Width=rgb->getRowSize();
			h->bV4Height=-r.ht; // negative ht indicates a "top-down" bitmap
			h->bV4Planes=1;
			h->bV4BitCount=8*sizeof(osl::graphics2d::RgbaPixel);
			h->bV4V4Compression=BI_BITFIELDS; //BI_RGB;
			h->bV4SizeImage=0; // Automatic
			h->bV4XPelsPerMeter=(long)(72.0/2.54*100.0); //Best guess: 72 dpi
			h->bV4YPelsPerMeter=(long)(72.0/2.54*100.0);
			h->bV4ClrUsed=0; //0; //All Colors used
			h->bV4ClrImportant=0; //All Colors needed
#if 0  //Storage: RGBA
			h->bV4RedMask  =0x000000ff;
			h->bV4GreenMask=0x0000ff00;
			h->bV4BlueMask =0x00ff0000;
			h->bV4AlphaMask=0xff000000;
#else  //Storage: BGRA (bugger-all)
			h->bV4RedMask  =0x00ff0000;
			h->bV4GreenMask=0x0000ff00;
			h->bV4BlueMask =0x000000ff;
			h->bV4AlphaMask=0xff000000;
#endif
			h->bV4CSType=LCS_WINDOWS_COLOR_SPACE;

			APICHECK(SetDIBitsToDevice,(hdc,x,y,
				r.wid,r.ht, 0,0, 0,r.ht, data,(BITMAPINFO *)h,DIB_RGB_COLORS));
		} else 
		{ //Convert to 24-bit format first
			BAD("Non-24 bit images not supported!\n");
		}
		ReleaseDC(hWnd,hdc);
	}
#endif
};
HINSTANCE myWindow::hInst; // current application instance
int myWindow::nCmdShow; // "show type" we were invoked with (e.g., run minimized)

/* Window procedure subroutine--feeds events to our class's wndProc */
LRESULT CALLBACK myWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	/* Find window class for the target winodw */
	myWindow *w=(myWindow *)GetWindowLongPtr(hWnd,GWLP_USERDATA);
	if (w!=NULL) 
		return w->classWndProc(message,wParam,lParam);
	else /* window not created yet?  or what? */
		return DefWindowProc(hWnd, message, wParam, lParam);
}

// Mesage handler for about box.
LRESULT CALLBACK myAboutBox(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	WT1("Entered about box message handler (%d)",message)
	switch (message)
	{
	case WM_INITDIALOG:
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) 
		{
			EndDialog(hDlg, LOWORD(wParam));
			return TRUE;
		}
		break;
	}
    return FALSE;
}

/** Parse command line into a set of arguments.  Currently unused. */
int parseCommandLine(char *args) {
	std::vector<std::string> argv; //Here's were we'll store the arguments as they come out
	
	//Locate the executable, to go in argv[0]
	char *exe=new char[MAX_PATH];
	::GetModuleFileName ( NULL, exe, MAX_PATH);
	argv.push_back(exe);
	delete[] exe;
	
	std::string curArg;
	while (*args!=0) {
		while(isspace(*args)) args++; //Skip over white space ahead of arguments
		if (*args==0) break; //That's the end of the arguments
		char quote=0;
		if (*args=='"' || *args=='\'')
			quote=*args++; //Strip off leading quote character
		int i;
		for (i=0;*args!=0;i++) {
			//Check if we're done yet
			if (quote) {
				if (*args==quote) { //End of quoted argument
					args++; //skip over quote character
					break; //Quote character terminates argument
				}
			}
			else /*no quote*/ {
				if (isspace(*args)) break; //Whitespace terminates unquoted argument
			}
			//If we get here, we're not done--keep copying the argument
			curArg+=*args++;
		}
		argv.push_back(curArg);
	}
	for (unsigned int v=0;v<argv.size();v++) WT2("Argument %d is [%s]",v,argv[v].c_str());
	//return main(argv.size(),&argv[0]);
	return 0;
}

/* Main program */
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	// Parse command line
	parseCommandLine(lpCmdLine);

	// Register window class
	myWindow::staticInit(hInstance,nCmdShow);

	// Create a window and process events for it
	myWindow *w=new myWindow();
	while (1) {w->waitEvent();}
    return 0;
}
